"""
This package provides convenient utilities and data to write a sphinx config file.
"""

from __future__ import annotations

from pathlib import Path
import json
from typing import Dict, Tuple, Set, Optional, cast

# See issue 4, we this the best format is Major.YYMM.day,
# in case of multiple releases a day we can borrow the next day's date.
# no 0 in front of the day as it is not pep440 compliant.
version_info = (0, 2510, 5)
__version__ = ".".join(map(str, version_info))

registry_file = Path(__file__).parent / "registry.json"

def _get_all_mappings() ->  Dict[str, Tuple[str, Optional[str]]]:
    return cast(
        Dict[str, Tuple[str, Optional[str]]],
        {k: tuple(v) for (k, v) in json.loads(registry_file.read_bytes()).items()},
    )


def get_intersphinx_mapping(
    *, packages: Set[str] = set()
) -> Dict[str, Tuple[str, Optional[str]]]:
    """
    Return values of intersphinx_mapping for sphinx configuration.

    For convenience, the returned dictionary is a copy so should be ok to
    mutate.

    Parameters
    ----------
    packages: Set of Str
        Libraries to include.

        Sphinx will download and load all the `objects.inv` for listed 
        packages. Getting all mappings is discourage as it will download all
        the `object.inv` which can be a non-negligible amount of data.

    """
    if len(packages) == 0:
        raise ValueError('You must explicitly give a list of packages for which to download intersphinx inventories: get_intersphinx_mapping(packages=["IPython", "numpy",...]).')

    mapping = _get_all_mappings()
    missing = set(packages) - set(mapping)
    if missing:
        raise ValueError(f"Some libraries in 'packages' not found in registry: {repr(sorted(missing))}")
    return {k: v for k, v in mapping.items() if k in packages}



import sys

from . import get_intersphinx_mapping
from . import __version__
from urllib.parse import urljoin

from typing import Optional

from sphinx.util.inventory import InventoryFile
from io import BytesIO

import requests

if len(sys.argv) not in [2, 3]:
    sys.exit(
        """Usage: python -m intersphinx_registry.lookup <package>[,package] [search_term]

        Example:

        $ python -m intersphinx_registry.lookup numpy,scipy array
        $ python -m intersphinx_registry.lookup ipython formatters.html

        """
    )


def main():
    print(f"Instersphinx-registry version {__version__}")

    packages = set(sys.argv[1].split(","))

    search_term: Optional[str]
    if len(sys.argv) == 3:
        search_term = sys.argv[2]
    else:
        search_term = None

    # there will be only one url
    urls = [
        (u[0], (u[1] if u[1] else "objects.inv"))
        for u in get_intersphinx_mapping(packages=packages).values()
    ]

    flattened = []
    for base_url, obj in urls:

        final_url = urljoin(base_url, obj)

        resp = requests.get(final_url)

        inv = InventoryFile.load(BytesIO(resp.content), base_url, urljoin)

        for key, v in inv.items():
            inv_entries = sorted(v.items())
            for entry, (_proj, _ver, url_path, display_name) in inv_entries:
                # display_name = display_name * (display_name != '-')
                flattened.append((key, entry, _proj, _ver, display_name, url_path))

    filtered = []

    width = [len(x) for x in flattened[0]]

    for item in flattened:
        key, entry, proj, version, display_name, url_path = item
        if (
            (search_term is None)
            or (search_term in entry)
            or (search_term in display_name)
            or (search_term in url_path)
        ):
            filtered.append((key, entry, proj, version, display_name, url_path))
            width = [max(w, len(x)) for w, x in zip(width, item)]


    for key, entry, proj, version, display_name, url_path in filtered:
        w_key, w_entry, w_proj, w_version, w_di, w_url = width
        print(
            f"{key:<{w_key}}  {entry:<{w_entry}}  {proj:<{w_proj}}  {version:<{w_version}}  {display_name!r:<{w_di+2}}  {url_path}"
        )


if __name__ == "__main__":
    main()



"""
Experimental, print info about all (or a subset of), all the known inventories.

As this is likely to make _many_ concurrent connections; 
we use aoihttp and sphobjinv which are not listed as dependencies.
"""

import asyncio
import sys
from urllib.parse import urljoin

import aiohttp
from sphobjinv import Inventory  # type: ignore

from intersphinx_registry import get_intersphinx_mapping

if len(sys.argv) not in [2, 3]:
    sys.exit(
        """Usage: python -m intersphinx_registry._info <package>[,package|:all]

        This prints debug info about registered package.

        WARNING: :all: will query _all_ the domains; this can be considered

        """
    )

packages = set(sys.argv[1].split(","))
if packages == {":all:"}:
    from intersphinx_registry import _get_all_mappings

    packages = set(list(_get_all_mappings().keys()))

print("Looking up metadata for", len(packages), "package")

# there will be only one url
urls = [
    (k, u[0], (u[1] if u[1] else "objects.inv"))
    for k, u in get_intersphinx_mapping(packages=packages).items()
]


async def fetch(session, key, base_url, obj):
    final_url = urljoin(base_url, obj)
    if obj != "objects.inv":
        print("Obj for ", base_url, obj)
    async with session.get(final_url) as response:
        return key, base_url, await response.read()


async def fetch_all(urls):
    async with aiohttp.ClientSession() as session:
        tasks = [fetch(session, *url) for url in urls]
        return await asyncio.gather(*tasks)


results = asyncio.run(fetch_all(urls))

flattened = []
for key, base_url, data in results:
    try:
        inv = Inventory(zlib=data)
    except Exception:
        print("ERROR with", key, base_url, file=sys.stderr)
        continue
    flattened.append((key, base_url, inv.project, inv.version, inv.count))

if not flattened:
    sys.exit("Could not reach any pacakges")
width = [len(x) for x in flattened[0][:-1]]

for item in flattened:
    width = [max(w, len(x)) for w, x in zip(width, item[:-1])]


for key, url, proj, version, count in flattened:
    w_key, w_url, w_proj, w_version = width
    print(
        f"{key:<{w_key}}  {proj!r:<{w_proj+2}} {version:<{w_version}} {count:<5} {url:<{w_url}}"
    )
